;+
; NAME: 
;  mkbias
; PURPOSE: 
;  Collect and combine CCD bias frames into a superbias frame
; DESCRIPTION:
;
;  The files are assumed to be named as 'root'NNN where 'root' is the
;  root of the file name that you supply and NNN is a three digit number.
;  If your file name has an imbedded '.' then add it to the root.
;
;  The specified range of files are all read in from FITS files.  Then each
;  image has the overscan mean subtracted (if desired), cropped (as indicated).
;  These images are then averaged.  The averaging is done with AVGCLIP.PRO
;  which does a robust average of the image stack so that cosmic rays or
;  other transient image defects are eliminated.
;
;  When done, the resulting bias image is returned to the caller and the image
;  is saved to a FITS file with the specified output filename.
;
; CATEGORY:
;  CCD data processing
; CALLING SEQUENCE:
;  mkbias,root,outsuf,start,nframes,bias
; INPUTS:
;  root    - Root of the file name (you must include . in the name).
;  outsuf  - The suffix of the final output file.
;  start   - First frame number to read (integer or long).
;               Start can also be a vector of explicit frame numbers to load.
;               In this case, nframes need not be specified and in fact will
;               be ignored.
;  nframes - Number of frames to average.
; OPTIONAL INPUT PARAMETERS:
; KEYWORD INPUT PARAMETERS:
;
;   CROP     = region of original image to save, default=no cropping.
;                 [x1,x2,y1,y2]
;
;   EXCLUDE - Optional vector of image numbers that should be excluded from
;                average.  Default is to include all frames.
;
;   OVERSCAN = column overscan region to use for frame bias level,
;                 default=no overscan subtraction.
;
; OUTPUTS:
;  bias - Final robust averaged/cropped bias image.
; KEYWORD OUTPUT PARAMETERS:
; COMMON BLOCKS:
; SIDE EFFECTS:
; RESTRICTIONS:
; PROCEDURE:
; MODIFICATION HISTORY:
;  95/03/09 - Initial crude version written, Marc W. Buie, Lowell Observatory
;  95/06/13, MWB, added OVERSCAN and CROP keywords
;  95/11/22, MWB, add EXCLUDE keyword
;  2000/02/02, MWB, rewrite to add support for multigroup FITS files.
;  2000/02/28, MWB, added support for frame numbers > 999.
;-
pro mkbias,root,outsuf,start,nframes,bias, $
       OVERSCAN=in_overscan,CROP=in_crop,EXCLUDE=exclude

   ;on_error,2

   if badpar(root,   7,       0,caller='MKBIAS: (root) '   ) then return
   if badpar(outsuf, 7,       0,caller='MKBIAS: (outsuf) ' ) then return
   if badpar(start,  [2,3],   [0,1],caller='MKBIAS: (start) ',rank=start_rank  ) then return

   if badpar(exclude,[0,2,3], [0,1],caller='MKBIAS: (exclude) ',default=-1) then return

   if badpar(in_overscan,[0,2,3],[1,2],caller='MKBIAS: (overscan) ',rank=o_rank) then return
   if badpar(in_crop,    [0,2,3],[1,2],caller='MKBIAS: (crop) ',rank=c_rank) then return

   ; Check to see if it's a sequential list or random list.
   if start_rank eq 0 then begin
      frames=start+indgen(nframes)
      if badpar(nframes,[2,3],   0,caller='MKBIAS: (nframes) ') then return
   endif else begin
      frames=start
      nframes=n_elements(frames)
   endelse

   ; Setup the file name format string
   digits = fix(ceil(alog10(max(frames)+1))) > 3
   dig = strn(digits)
   fnfmt = '(i'+dig+'.'+dig+')'

   ; Apply the exclusion criteria to the frames list.  Then, find out how
   ;   many frames are being collected and make sure there is something to do.
   for i=0,nframes-1 do begin
      z=where(frames[i] eq exclude,count)
      if count ne 0 then frames[i]=-1
   endfor
   zg=where(frames ne -1,countg)
   if countg eq 0 then $
      message,'Error ** you have excluded all frames, nothing to do.'

   ; Make the first file name
   fname=root+string(frames[0],format=fnfmt)
   if exists(fname+'.fits') then fname=fname+'.fits'

   ; Make the output file name
   outfile = root+outsuf

   ; Check header of image to see if it is a multi-extension image.
   hdr=headfits(fname)
   numext=sxpar(hdr,'NEXTEND')

   ; Setup the overscan/crop control values
   if numext eq 0 then begin
      extend=0
      numext=1
      if o_rank eq 0 then begin
         do_overscan=0
      endif else begin
         do_overscan=1
         overscan = in_overscan
      endelse
      if c_rank eq 0 then begin
         do_crop=0
      endif else begin
         do_crop=1
         crop = in_crop
      endelse
   endif else begin
      extend=1
      if o_rank eq 0 then begin
         do_overscan=0
      endif else if o_rank eq 1 then begin
         do_overscan=1
         overscan = rebin(in_overscan,n_elements(overscan),numext)
      endif else begin
         do_overscan=1
         overscan = in_overscan
      endelse
      if c_rank eq 0 then begin
         do_crop=0
      endif else if o_rank eq 1 then begin
         do_crop=1
         crop = rebin(in_crop,n_elements(crop),numext)
      endif else begin
         do_crop=1
         crop = in_crop
      endelse
   endelse

   ; If it's multi-extension, then the header we've just read is special
   ;   And must be written back out to start the output file.
   if extend then $
      writefits,outfile,0,hdr

   ; Main loop over extension, done just once on "normal" frames
   for ix=1,numext do begin
      ix0=ix-1

      if extend then $
         print,'Load frames ',strn(min(frames)),' to ', $
                              strn(max(frames)),' extension ',strn(ix) $
      else $
         print,'Load frames ',strn(min(frames)),' to ',strn(max(frames))

      ; Setup the first file name.
      fname=root+string(frames[zg[0]],format=fnfmt)
      if exists(fname+'.fits') then fname=fname+'.fits'
      if not exists(fname) then begin
         print,'MKBIAS:  ',fname,' not found, unable to continue.'
         return
      endif

      ; Loop over the frames
      for i=0,nframes-1 do begin

         fname=root+string(frames[zg[i]],format=fnfmt)
         if exists(fname+'.fits') then fname=fname+'.fits'

         ; read in the new image, the dummy line before readfits is to dump
         ;   the storage for the array before it's used again.
         image = 0
         if i eq 0 then begin
            if extend then $
               image = readfits(fname,hdr,exten_no=ix,/silent) $
            else $
               image = readfits(fname,hdr,/silent)
         endif else begin
            if extend then $
               image = readfits(fname,exten_no=ix,/silent) $
            else $
               image = readfits(fname,/silent)
         endelse

         if do_overscan and do_crop then begin
            image = colbias(image,overscan[0,ix0],overscan[1,ix0], $
                     crop[0,ix0],crop[1,ix0],crop[2,ix0],crop[3,ix0],biasval=biasval)
            if extend then $
               print,fname,' overscan value is ',biasval,' extension ',strn(ix) $
            else $
               print,fname,' overscan value is ',biasval

         endif else if do_overscan then begin
            image = colbias(image,overscan[0,ix0],overscan[1,ix0],biasval=biasval)
            if extend then $
               print,fname,' overscan value is ',biasval,' extension ',strn(ix) $
            else $
               print,fname,' overscan value is ',biasval

         endif else if do_crop then begin
            image = image[crop[0,ix0]:crop[1,ix0],crop[2,ix0]:crop[3,ix0]]
            if extend then $
               print,fname,' extension ',strn(ix) $
            else $
               print,fname

         endif else begin
            if extend then $
               print,fname,' extension ',strn(ix) $
            else $
               print,fname

         endelse

         if i eq 0 then begin
            sz=size(image,/dim)
            cube=fltarr(sz[0],sz[1],countg,/nozero)
         endif

         cube[*,*,i]=image

      endfor

      ; Average the cube down.
      bias=0
      avgclip,cube,bias

      sxaddpar,hdr,'NAXIS1',sz[0]
      sxaddpar,hdr,'NAXIS2',sz[1]
      sxaddpar,hdr,'BITPIX',-32
      if extend then begin
         print,outfile,',  writing extension ',strn(ix)
         writefits,outfile,bias,hdr,/append
      endif else begin
         print,'Saving final bias frame to ',outfile
         writefits,outfile,bias,hdr
      endelse

   endfor

end
