;+
; NAME: 
;  mkdark
; PURPOSE: 
;  Collect and combine CCD dark frames into a superdark frame
; DESCRIPTION:
;
;  The files are assumed to be named as 'root'NNN where 'root' is the
;  root of the file name that you supply and NNN is a three digit number.
;  If your file name has an imbedded '.' then add it to the root.
;
;  The specified range of files are all read in from FITS files.  Then each
;  image has the overscan mean subtracted (if desired), cropped (as indicated),
;  and then the input bias image is subtracted.  These images are then
;  normalized to counts per second and then averaged.  The averaging
;  is done with AVGCLIP.PRO which does a robust average of the image
;  stack so that cosmic rays or other transient image defects are eliminated.
;
;  When done, the resulting dark image is returned to the caller and the image
;  is saved to a FITS file with the specified output filename.
;
; CATEGORY:
;  CCD data processing
; CALLING SEQUENCE:
;  mkdark,root,outsuf,start,nframes,bias,dark
; INPUTS:
;  root    - Root of the file name (you must include . in the name).
;  outsuf  - The suffix of the final output file.
;  start   - First frame number to read (integer or long).
;  nframes - Number of frames to average.
;  bias    - Bias frame image to subtract from each raw dark.
; OPTIONAL INPUT PARAMETERS:
; KEYWORD INPUT PARAMETERS:
;
;   CROP     = region of original image to save, default=no cropping.
;                 [x1,x2,y1,y2]
;
;   EXCLUDE - Optional vector of image numbers that should be excluded from
;                average.  Default is to include all frames.
;
;   EXPKEY = String - FITS keyword to read to get exposure time, default = EXPTIME
;
;   OVERSCAN = column overscan region to use for frame bias level,
;                 default=no overscan subtraction.
;
; OUTPUTS:
;  dark - Final robust averaged dark image scaled to counts per second.
; KEYWORD OUTPUT PARAMETERS:
; COMMON BLOCKS:
; SIDE EFFECTS:
; RESTRICTIONS:
; PROCEDURE:
; MODIFICATION HISTORY:
;  95/03/09 - Initial crude version written, Marc W. Buie, Lowell Observatory
;  95/08/30, MWB, added OVERSCAN and CROP keywords
;  95/11/22, MWB, add EXCLUDE keyword
;  99/06/10, MWB, added EXPTIME keyword and added documentation.
;-
pro mkdark,root,outsuf,start,nframes,bias,dark, $
       OVERSCAN=overscan,CROP=crop,EXCLUDE=exclude,EXPKEY=expkey

on_error,2

if badpar(root,    7,        0,caller='MKDARK: (root) '    ) then return
if badpar(outsuf,  7,        0,caller='MKDARK: (outsuf) '  ) then return
if badpar(start,   [2,3],    0,caller='MKDARK: (start) '   ) then return
if badpar(nframes, [2,3],    0,caller='MKDARK: (nframes) ' ) then return
if badpar(bias,    [2,3,4,5],2,caller='MKDARK: (bias) '    ) then return
if badpar(overscan,[0,2,3],  1,caller='MKDARK: (overscan) ') then return
if badpar(exclude,[0,2,3],[0,1],caller='MKDARK: (exclude) ',default=-1) then return
if badpar(crop,    [0,2,3],  1,caller='MKDARK: (crop) '    ) then return
if badpar(expkey,  [0,7],    0,caller="MKDARK: (EXPKEY) ",default="EXPTIME") then return

frames=start+indgen(nframes)
print,'Load frames ',start,' to ',start+nframes-1
for i=0,nframes-1 do begin
   z=where(frames[i] eq exclude,count)
   if count ne 0 then frames[i]=-1
endfor
z=where(frames ne -1,count)
if count eq 0 then $
   message,'Error ** you have excluded all frames, nothing to do.'

fname=root+string(frames[z[0]],format='(i3.3)')
first = readfits(fname,hdr,/silent)
exptime=sxpar(hdr,expkey)
sz=size(first)
firstframe=frames[z[0]]

if keyword_set(overscan) and keyword_set(crop) then begin
   firstn = colbias(first,overscan[0],overscan[1], $
               crop[0],crop[1],crop[2],crop[3],biasval=biasval)
   sz=size(firstn)
   cube=fltarr(sz[1],sz[2],count,/nozero)
   cube[*,*,0]=(firstn-bias)/float(exptime)
   print,fname,exptime,biasval
   j=1
   for i=1,nframes-1 do begin
      if frames[i] ne -1 and frames[i] ne firstframe then begin
         fname=root+string(frames[i],format='(i3.3)')
         frame = readfits(fname,hdr,/silent)
         exptime=sxpar(hdr,expkey)
         cube[*,*,j] = (colbias(frame,overscan[0],overscan[1], $
                           crop[0],crop[1],crop[2],crop[3], $
                           biasval=biasval) - $
                                   bias ) / float(exptime)
         print,fname,exptime,biasval
         j=j+1
      endif
   endfor
endif else if keyword_set(overscan) then begin
   firstn = colbias(first,overscan[0],overscan[1],biasval=biasval)
   sz=size(firstn)
   cube=fltarr(sz[1],sz[2],count,/nozero)
   cube[*,*,0]=(firstn-bias)/float(exptime)
   print,fname,exptime,biasval
   j=1
   for i=1,nframes-1 do begin
      if frames[i] ne -1 and frames[i] ne firstframe then begin
         fname=root+string(frames[i],format='(i3.3)')
         frame = readfits(fname,hdr,/silent)
         exptime=sxpar(hdr,expkey)
         cube[*,*,j] = (colbias(frame,overscan[0],overscan[1],biasval=biasval) - $
                                 bias ) / float(exptime)
         print,fname,exptime,biasval
         j=j+1
      endif
   endfor
endif else if keyword_set(crop) then begin
   firstn = first[crop[0]:crop[1],crop[2]:crop[3]]
   sz=size(firstn)
   cube=fltarr(sz[1],sz[2],count,/nozero)
   cube[*,*,0]=(firstn-bias)/float(exptime)
   print,fname,exptime
   j=1
   for i=1,nframes-1 do begin
      if frames[i] ne -1 and frames[i] ne firstframe then begin
         fname=root+string(frames[i],format='(i3.3)')
         frame = readfits(fname,hdr,/silent)
         exptime=sxpar(hdr,expkey)
         cube[*,*,j] = (frame[crop[0]:crop[1],crop[2]:crop[3]]-bias)/float(exptime)
         print,fname,exptime
         j=j+1
      endif
   endfor
endif else begin
   cube=fltarr(sz[1],sz[2],count,/nozero)
   cube[*,*,0]=(first-bias)/float(exptime)
   print,fname,exptime
   j=1
   for i=1,nframes-1 do begin
      if frames[i] ne -1 and frames[i] ne firstframe then begin
         fname=root+string(frames[i],format='(i3.3)')
         cube[*,*,j] = readfits(fname,hdr,/silent)-bias
         exptime=sxpar(hdr,expkey)
         cube[*,*,j] = cube[*,*,j] / float(exptime)
         print,fname,exptime
         j=j+1
      endif
   endfor
endelse

avgclip,cube,dark
writefits,root+outsuf,dark,hdr
print,'Saving final dark frame to ',root+outsuf

end
